DungeonMate = LibStub("AceAddon-3.0"):NewAddon("DungeonMate","AceConsole-3.0","AceEvent-3.0")

local version = 0.1

local dm_panel, plist

-- prototyped functions
local clearParty, insertPlayer

local roleicons = "Interface\\LFGFrame\\UI-LFG-ICON-PORTRAITROLES.blp"

local function IgnorePlayer(frame)
	AddOrDelIgnore(frame.player_name)
end

local function GenerateGroupInfo()
	local result = {}
	local n = GetNumPartyMembers()
	while n > 0 do
		local unit = "party" .. n
		local name, realm = UnitName(unit)
		local isTank, isHeal, isDPS = UnitGroupRolesAssigned(unit)
		local classname,class = UnitClass(unit)
		if realm then name = name .. "-" .. realm end
		result[n] = {["name"] = name}
		if isTank then
			result[n].role = "tank"
		elseif isHeal then
			result[n].role = "heal"
		elseif isDPS then
			result[n].role = "dps"
		else
			result[n].role = "no role"
		end
		result[n].class = class
		result[n].classname = classname
		n = n - 1
	end
	--[[ TEST DATA
	result = { [1]={["name"]="Bert",
	                ["role"]="heal",
	                ["class"]="PRIEST"
	               },
	           [2]={["name"]="Ernie",
	                ["role"]="dps",
	                ["class"]="WARLOCK"
	               },
	           [3]={["name"]="Snuffleupagus",
	                ["role"]="tank",
	                ["class"]="DEATHKNIGHT"
	               },
	           [4]={["name"]="Bigbird",
	                ["role"]="dps",
	                ["class"]="SHAMAN"
	               }
	         }
	]]--
	return result
end

local function updatePlayer(player, frame)
	frame.info = player -- export the original info, in case it helps anyone.
	frame.label:SetText(player.name)
	frame.ignorebutton.player_name = player.name -- Button Kludge FIXME
	
	if player.class == "DEATHKNIGHT" then
		frame.label:SetTextColor(0.77,0.12,0.23)
	elseif player.class == "DRUID" then
		frame.label:SetTextColor(1.00,0.49,0.04)
	elseif player.class == "HUNTER" then
		frame.label:SetTextColor(0.67,0.83,0.45)
	elseif player.class == "MAGE" then
		frame.label:SetTextColor(0.41,0.80,0.94)
	elseif player.class == "PALADIN" then
		frame.label:SetTextColor(0.96,0.55,0.73)
	elseif player.class == "PRIEST" then
		frame.label:SetTextColor(1.00,1.00,1.00)
	elseif player.class == "ROGUE" then
		frame.label:SetTextColor(1.00,0.96,0.41)
	elseif player.class == "SHAMAN" then
		frame.label:SetTextColor(0.14,0.35,1.00)
	elseif player.class == "WARLOCK" then
		frame.label:SetTextColor(0.58,0.51,0.79)
	elseif player.class == "WARRIOR" then
		frame.label:SetTextColor(0.78,0.61,0.43)
	else 
		frame.label:SetTextColor(0.5,0.5,0.5)
	end
	
	if player.role == "tank" then
		frame.rolebutton:SetTexCoord(0, 19/64, 22/64, 41/64)
		frame.rolebutton:Show()
	elseif player.role == "heal" then
		frame.rolebutton:SetTexCoord(20/64, 39/64, 1/64, 20/64)
		frame.rolebutton:Show()
	elseif player.role == "dps" then
		frame.rolebutton:SetTexCoord(20/64, 39/64, 22/64, 41/64)
		frame.rolebutton:Show()
	else
		frame.rolebutton:Hide()
	end
end

local function UpdatePartyDisplay()
	if not plist then return end
	local group = GenerateGroupInfo()
	local k,v
	for _,v in ipairs(group) do
		local found = false
		-- Next, loop over the already produced frames to see if we've seen this unit already.
		-- NOTE: This is a fairly poor algorithm which operates O(n^2) in the worst case. It could
		-- be improved upon with more infrastructure, but frankly the coefficients are small enough
		-- that it doesn't much matter.
		for _,k in ipairs(plist.group) do
			if k.info.name == v.name then 
				found = true
				updatePlayer(v,k)
				break
			end
		end
		if not found then
			insertPlayer(v)
		end
	end
end

local function ResetDisplay()
	clearParty()
	UpdatePartyDisplay()
end

local function showPanel()
	UpdatePartyDisplay()
	dm_panel:Show()
end

local function hidePanel()
	dm_panel:Hide()
end

local function saveFramePositions()
		local point,rt,rp,xoff,yoff = plist:GetPoint(1)
		DungeonMate.db.char.panel_position.x = xoff
		DungeonMate.db.char.panel_position.y = yoff
		DungeonMate.db.char.panel_position.point = point
end

local function initializeFrame()
	dm_panel = CreateFrame("Frame",nil,UIParent)
	dm_panel:SetBackdrop({bgFile ="Interface/DialogFrame/UI-DialogBox-Background", 
						  edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
						  tile = true, tileSize = 16, edgeSize = 16, 
						  insets = { left = 4, right = 4, top = 4, bottom = 4 }})

	
	local title = dm_panel:CreateFontString(nil,"OVERLAY","GameFontHighlightLarge")
	title:SetPoint("TOPLEFT",dm_panel,15,-11)
	title:SetText("DungeonMate v" .. version)
	title:Show()
	
	local footer = CreateFrame("Frame",nil,dm_panel)
	footer:SetHeight(25)
	footer:Show()
	
	-- This button used to say "Update", hence the name. 
	local update = CreateFrame("Button",nil,footer,"UIPanelButtonTemplate")
	update:SetText("Reset")
	update:SetScript("OnClick",ResetDisplay)
	update:Show()
	
	local close = CreateFrame("Button",nil,dm_panel,"UIPanelCloseButton")
	close:SetPoint("TOPRIGHT",dm_panel,"TOPRIGHT")
	close:SetScript("OnClick",hidePanel)

	plist = CreateFrame("Frame",nil,dm_panel)
	plist:SetPoint(DungeonMate.db.char.panel_position.point,UIParent,
	               DungeonMate.db.char.panel_position.point,
	               DungeonMate.db.char.panel_position.x,
	               DungeonMate.db.char.panel_position.y)
	plist:SetWidth(250)
	plist:SetHeight(4) -- dummy height, changes with added players
	plist.group = {}
	plist:Show()
	
	
	footer:SetPoint("BOTTOMLEFT",plist,"BOTTOMLEFT",0,-1*(footer:GetHeight()+4))
	footer:SetPoint("BOTTOMRIGHT",plist,"BOTTOMRIGHT",0,-1*(footer:GetHeight()+4))
	update:SetPoint("TOPLEFT",footer,"TOPLEFT",50,0)
	update:SetPoint("BOTTOMRIGHT",footer,"BOTTOMRIGHT",-50,0)
	dm_panel:SetPoint("TOPLEFT",plist,"TOPLEFT",-10,30)
	dm_panel:SetPoint("BOTTOMRIGHT",footer,"BOTTOMRIGHT",10,-5)
	
	dm_panel:EnableMouse(true)
	plist:SetMovable(true)
	dm_panel:SetScript("OnMouseUp",function(frame)
		plist:StopMovingOrSizing()
		saveFramePositions()
	end)
	dm_panel:SetScript("OnMouseDown",function(frame,button)
		if button == "LeftButton" then
			plist:StartMoving()
		end
	end)
	dm_panel:SetScript("OnHide",function(frame)
		plist:StopMovingOrSizing()
		saveFramePositions()
	end)
	
	dm_panel:Show()
end

local playerRecycleQueue = {}

local function recyclePlayer(player)
	player:Hide()
	tinsert(playerRecycleQueue,player)
end

local function manufacturePlayerFrame(player)
	local resultframe -- Input player is a player info table. Output is a frame based on that.
	if #playerRecycleQueue > 0 then
		resultframe = tremove(playerRecycleQueue) -- ok, so it's a stack... big deal
	else
		resultframe = CreateFrame("Frame",nil,plist)
		resultframe.label = resultframe:CreateFontString(nil,"OVERLAY","GameFontNormal")
		resultframe:SetHeight(25)
		resultframe.label:SetPoint("TOPLEFT",resultframe,"TOPLEFT",20,0)
		resultframe.label:SetPoint("BOTTOMLEFT",resultframe,"BOTTOMLEFT",20,0)
		resultframe.rolebutton = resultframe:CreateTexture(nil,"OVERLAY")
		resultframe.rolebutton:SetPoint("LEFT",resultframe,"LEFT",0,0)
		resultframe.rolebutton:SetTexture(roleicons)
		resultframe.rolebutton:SetHeight(16)
		resultframe.rolebutton:SetWidth(16)
		resultframe.ignorebutton = CreateFrame("Button",nil,resultframe,"UIPanelButtonTemplate")
		resultframe.ignorebutton:SetPoint("TOPRIGHT",resultframe,"TOPRIGHT")
		resultframe.ignorebutton:SetPoint("BOTTOMRIGHT",resultframe,"BOTTOMRIGHT")
		resultframe.ignorebutton:SetWidth(60)
		resultframe.ignorebutton:SetText("Ignore")
		resultframe.ignorebutton:SetScript("OnClick",IgnorePlayer)
		resultframe.ignorebutton:Show();
	end
	
	updatePlayer(player,resultframe)
	
	resultframe:Show()
	return resultframe
end

-- PROTOTYPED as local
function clearParty()
	if not plist then return end
	local temp = tremove(plist.group)
	while temp do
		temp:Hide()
		plist:SetHeight(plist:GetHeight()-temp:GetHeight())
		recyclePlayer(temp)
		temp = tremove(plist.group)
	end
end

-- PROTOTYPED as local
function insertPlayer(player)
	-- Keep in mind, at this point "player" is an info table, not a frame
	if not plist then return end
	local frame = manufacturePlayerFrame(player)
	if #(plist.group) > 0 then
		-- Attach to the last player
		local last = plist.group[#(plist.group)]
		frame:SetPoint("TOPLEFT",last,"BOTTOMLEFT")
		frame:SetPoint("TOPRIGHT",last,"BOTTOMRIGHT")
	else
		-- Attach to the top of the player list
		frame:SetPoint("TOPLEFT",plist,"TOPLEFT")
		frame:SetPoint("TOPRIGHT",plist,"TOPRIGHT")
	end
	plist:SetHeight(plist:GetHeight()+frame:GetHeight())
	tinsert(plist.group,frame)
end

function DungeonMate:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("DungeonMateDB")
	self.db:RegisterDefaults({
		char = {
			["panel_position"] = {
				["point"] = "CENTER",
				["x"] = 450,
				["y"] = 0,
			}
		}
	})
	self:RegisterChatCommand("dm",showPanel)
	self:RegisterChatCommand("dungeonmate",showPanel)
end

function DungeonMate:OnEnable()
	initializeFrame()
	showPanel()
	DungeonMate:RegisterEvent("PARTY_MEMBERS_CHANGED",UpdatePartyDisplay)
end

function DungeonMate:OnDisable()
	hidePanel()
end

